"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CertificateAggregate = void 0;
const Certificate_events_1 = require("./events/Certificate.events");
const errors_1 = require("./errors");
const ethers_1 = require("ethers");
const date_utils_1 = require("../utils/date-utils");
class CertificateAggregate {
    constructor(events) {
        const sortedEvents = [...events].sort((a, b) => date_utils_1.compareDates(a.createdAt, b.createdAt));
        const [firstEvent, ...restEvents] = sortedEvents;
        if (!firstEvent) {
            throw new errors_1.CertificateErrors.CertificateNoEvents();
        }
        if (firstEvent.type !== Certificate_events_1.CertificateEventType.Issued) {
            throw new errors_1.CertificateErrors.FirstCertificateEventIsNotIssuance(firstEvent.internalCertificateId, firstEvent.type);
        }
        const initialCertificate = this.createCertificate(firstEvent);
        this.certificate = {
            ...restEvents.reduce(this.apply.bind(this), initialCertificate),
            isSynced: this.reduceIsSynced(events)
        };
    }
    apply(certificate, event) {
        switch (event.type) {
            case Certificate_events_1.CertificateEventType.Issued:
                return this.issue(certificate, event);
            case Certificate_events_1.CertificateEventType.Transferred:
                return this.transfer(certificate, event);
            case Certificate_events_1.CertificateEventType.Claimed:
                return this.claim(certificate, event);
            case Certificate_events_1.CertificateEventType.ClaimPersisted:
                return this.claimPersisted(certificate, event);
            case Certificate_events_1.CertificateEventType.IssuancePersisted:
                return this.issuancePersisted(certificate, event);
            case Certificate_events_1.CertificateEventType.TransferPersisted:
                return this.transferPersisted(certificate, event);
            case Certificate_events_1.CertificateEventType.PersistError:
                return this.persistError(certificate, event);
            default:
                throw new errors_1.CertificateErrors.UnknownEventType(event.internalCertificateId, event.type);
        }
    }
    getCertificate() {
        return this.certificate;
    }
    static fromEvents(events) {
        return new CertificateAggregate(events);
    }
    createCertificate(event) {
        const { payload } = event;
        return {
            internalCertificateId: event.internalCertificateId,
            blockchainCertificateId: null,
            claims: [],
            creationBlockHash: '',
            creationTime: Math.floor(Date.now() / 1000),
            deviceId: payload.deviceId,
            generationStartTime: payload.fromTime,
            generationEndTime: payload.toTime,
            metadata: payload.metadata,
            owners: {
                [payload.toAddress]: payload.energyValue
            },
            claimers: {},
            isSynced: false,
            transactions: []
        };
    }
    persistError(certificate, event) {
        return certificate;
    }
    issue(certificate, event) {
        throw new errors_1.CertificateErrors.Issuance.CertificateAlreadyIssued(event.internalCertificateId);
    }
    issuancePersisted(certificate, event) {
        const { blockchainCertificateId } = event.payload;
        return {
            ...certificate,
            blockchainCertificateId,
            transactions: [
                ...certificate.transactions,
                {
                    eventType: Certificate_events_1.CertificateEventType.IssuancePersisted,
                    timestamp: event.createdAt,
                    transactionHash: event.payload.transactionHash
                }
            ]
        };
    }
    transfer(certificate, event) {
        this.validateTransfer(certificate, event);
        const { fromAddress, toAddress, energyValue } = event.payload;
        const fromBalance = ethers_1.BigNumber.from(certificate.owners[fromAddress] ?? '0');
        const toBalance = ethers_1.BigNumber.from(certificate.owners[toAddress] ?? '0');
        const newFromBalance = fromBalance.sub(energyValue);
        const newToBalance = toBalance.add(energyValue);
        return {
            ...certificate,
            owners: {
                ...certificate.owners,
                [fromAddress]: newFromBalance.toString(),
                [toAddress]: newToBalance.toString()
            }
        };
    }
    transferPersisted(certificate, event) {
        return {
            ...certificate,
            transactions: [
                ...certificate.transactions,
                {
                    eventType: Certificate_events_1.CertificateEventType.TransferPersisted,
                    timestamp: event.createdAt,
                    transactionHash: event.payload.transactionHash
                }
            ]
        };
    }
    claim(certificate, event) {
        this.validateClaim(certificate, event);
        const { forAddress, energyValue, claimData } = event.payload;
        const unclaimedBalance = ethers_1.BigNumber.from(certificate.owners[forAddress]);
        const claimedBalance = ethers_1.BigNumber.from(certificate.claimers[forAddress] ? certificate.claimers[forAddress] : '0');
        const newUnclaimedBalance = unclaimedBalance.sub(energyValue);
        const newClaimedBalance = claimedBalance.add(energyValue);
        return {
            ...certificate,
            owners: {
                ...certificate.owners,
                [forAddress]: newUnclaimedBalance.toString()
            },
            claimers: {
                ...certificate.claimers,
                [forAddress]: newClaimedBalance.toString()
            },
            claims: [
                ...certificate.claims,
                {
                    id: event.internalCertificateId,
                    from: forAddress,
                    to: forAddress,
                    topic: '',
                    value: energyValue,
                    claimData: claimData
                }
            ]
        };
    }
    claimPersisted(certificate, event) {
        return {
            ...certificate,
            transactions: [
                ...certificate.transactions,
                {
                    eventType: Certificate_events_1.CertificateEventType.ClaimPersisted,
                    timestamp: event.createdAt,
                    transactionHash: event.payload.transactionHash
                }
            ]
        };
    }
    validateTransfer(certificate, event) {
        const { internalCertificateId, payload: { fromAddress, toAddress, energyValue } } = event;
        if (this.isZeroAddress(fromAddress)) {
            throw new errors_1.CertificateErrors.Transfer.FromZeroAddress(internalCertificateId);
        }
        if (this.isZeroAddress(toAddress)) {
            throw new errors_1.CertificateErrors.Transfer.ToZeroAddress(internalCertificateId);
        }
        if (!this.hasEnoughBalance(certificate, fromAddress, energyValue)) {
            throw new errors_1.CertificateErrors.Transfer.NotEnoughBalance(internalCertificateId, fromAddress);
        }
    }
    validateClaim(certificate, event) {
        const { internalCertificateId, payload: { forAddress, energyValue } } = event;
        if (this.isZeroAddress(forAddress)) {
            throw new errors_1.CertificateErrors.Claim.ForZeroAddress(internalCertificateId);
        }
        if (!this.hasEnoughBalance(certificate, forAddress, energyValue)) {
            throw new errors_1.CertificateErrors.Claim.NotEnoughBalance(internalCertificateId, forAddress);
        }
    }
    isZeroAddress(address) {
        return parseInt(address, 16) === 0;
    }
    hasEnoughBalance(certificate, address, volume) {
        const addressBalance = certificate.owners[address];
        return (addressBalance !== undefined &&
            ethers_1.BigNumber.from(addressBalance).gte(ethers_1.BigNumber.from(volume)));
    }
    reduceIsSynced(events) {
        const toPersistCounter = events.reduce((toPersistCounter, event) => {
            switch (event.type) {
                case Certificate_events_1.CertificateEventType.Issued:
                case Certificate_events_1.CertificateEventType.Transferred:
                case Certificate_events_1.CertificateEventType.Claimed:
                    return toPersistCounter + 1;
                case Certificate_events_1.CertificateEventType.ClaimPersisted:
                case Certificate_events_1.CertificateEventType.IssuancePersisted:
                case Certificate_events_1.CertificateEventType.TransferPersisted:
                    return toPersistCounter - 1;
                case Certificate_events_1.CertificateEventType.PersistError:
                    return toPersistCounter;
                default:
                    throw new errors_1.CertificateErrors.UnknownEventType(event.internalCertificateId, event.type);
            }
        }, 0);
        if (toPersistCounter > 0) {
            return false;
        }
        else if (toPersistCounter === 0) {
            return true;
        }
        else {
            throw new errors_1.CertificateErrors.CertificateTooManyPersisted(events[0].internalCertificateId, Math.abs(toPersistCounter));
        }
    }
}
exports.CertificateAggregate = CertificateAggregate;
//# sourceMappingURL=certificate.aggregate.js.map